/* 
	MIDI to CV converter main

	$RCSfile$
	$Revision: 143 $
	$Author  Satoshi "Chuck" Takagi $
	$Date: 2008-12-30 06:22:53 +0900 (火, 30 12 2008) $
 */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include "global.h"
#include "uart.h"
#include "midi.h"
#include "avrport.h"
#include "config.h"

struct _MidiNote {
	uint8_t Note;
	uint8_t Velocity;
	struct _MidiNote *next;
};
struct _MidiNote NoteEnd = { 255, 255, 0 };
struct _MidiNote NoteStart = { 0, 0, &NoteEnd };
struct _MidiNote MidiNotes[LEGATO_MEMSIZ];
static uint16_t MidiNotesTag;

/*
uint8_t mcusr_mirror;
void get_mcusr(void) \
__attribute__((naked)) \
__attribute__((section(".init3")));

void get_mcusr(void)
{
	mcusr_mirror = MCUSR;
	MCUSR = 0;
	wdt_disable();
}
*/

uint8_t g_midi_ch;

static uint8_t g_expvel;
static uint8_t g_cv0set;
static uint8_t g_cv1set;
static uint8_t g_cv2set;
static uint8_t g_cv3set;
static uint8_t g_g0set;
static uint8_t g_g1set;
static uint8_t g_g2set;
static uint8_t g_g3set;

#ifdef MCV2313A
static uint8_t g_note, g_vel, g_pitch;
#endif // MCV2313A

static uint8_t g_cc0, g_cc1, g_cc2, g_cc3;

volatile static uint8_t g_gate = 0;
volatile static uint8_t legato_time = 0;	// no retrigger default
volatile static uint8_t retrigger_time = 0;
volatile static uint8_t legato_depth = LEGATO_MEMSIZ;

volatile static uint8_t tmcnt1m;

ISR(TIMER1_OVF_vect)
{
	// 256 cnt in 20MHz system
	// 256/20M * 78 = 1msec
	if (tmcnt1m)
		tmcnt1m--;
	else {
		tmcnt1m = 77;
		if (retrigger_time)
			retrigger_time--;
	}
}


static void ioinit (void)
{
	uint8_t tmpb, tmpd;
	PORTB = 0xFF;
	DDRB = (1<<DDB0)|(1<<DDB1)|(1<<DDB2)|(1<<DDB3)|(1<<DDB4);
	PORTD = 0x7E;
	DDRD = (1<<DDD4)|(1<<DDD5)|(1<<DDD6);

	// input settings
	// MIDI CH3210 = PB5, PD3, PD2, PD1
	// CV2SEL = PB6, CV3SEL = PB7
	tmpb = ~PINB;
	tmpd = ~PIND;
	g_midi_ch = ((tmpb>>2)&0x08)|((tmpd>>1)&0x07);

	g_cv0set = MIDI_CONFIG_CV0_src;
	g_cv1set = MIDI_CONFIG_CV1_src;
	g_g0set = MIDI_CONFIG_GATE0_src;
	g_g1set = MIDI_CONFIG_GATE1_src;
	g_g2set = MIDI_CONFIG_GATE2_src;
	if (tmpb & 0x40) {
		g_cv2set = MIDI_CONFIG_CV2_alt_src;
		g_expvel = 1;	// 1: Alt, 0: src
	} else {
		g_cv2set = MIDI_CONFIG_CV2_src;
	}

	if (tmpb & 0x80) {
		g_cv3set = MIDI_CONFIG_CV3_alt_src;
		g_g3set = MIDI_CONFIG_GATE3_alt_src;
	} else {
		g_cv3set = MIDI_CONFIG_CV3_src;
		g_g3set = MIDI_CONFIG_GATE3_src;
	}

	
	uart_init();

	// Fast PWM 8bit
	TCCR0A = (1<<COM0A1)|(1<<COM0A0)|(1<<COM0B1)|(1<<COM0B0)|(1<<WGM01)|(1<<WGM00);
	TCCR0B = (1<<CS00);
	TCCR1A = (1<<COM1A1)|(1<<COM1A0)|(1<<COM1B1)|(1<<COM1B0)|(1<<WGM10);
	TCCR1B = (1<<WGM12)|(1<<CS10);
	TIMSK = (1<<TOIE1);
	tmcnt1m = 77;

	CV0 = CVCALC(0x00);
	CV1 = CVCALC(0x00);
	CV2 = CVCALC(0x00);
	CV3 = CVCALC(0x00);
	GATE0OFF;
	GATE1OFF;
	GATE2OFF;
	GATE3OFF;

//	wdt_enable(WDTO_4S);
}


static void midi_cvgate(void)
{
	if (MidiChanMes.ControlChange.Control == MIDI_CC_LEGATO) {
		if (MidiChanMes.ControlChange.Value < 64) {
			// Legato Off // same processing as all note off
			g_gate = 0;
			MidiNotesTag = 0x0000;
			NoteEnd.next = NULL;
			NoteStart.next = &NoteEnd;
			legato_time = MidiChanMes.ControlChange.Value;
			legato_depth = 1;
		} else {
			// Legato On
#ifdef RETRIGGER_TIME_ORDER_DEC
			legato_time = 127 - MidiChanMes.ControlChange.Value;
#else
			legato_time = MidiChanMes.ControlChange.Value - 64;
#endif // RETRIGGER_TIME_ORDER_DEC
			legato_depth = LEGATO_MEMSIZ;
		}
	}

	if (MidiChanMes.ControlChange.Control == g_cv0set)
		g_cc0 = CVCALC(MidiChanMes.ControlChange.Value);
	if (MidiChanMes.ControlChange.Control == g_cv1set)
		g_cc1 = CVCALC(MidiChanMes.ControlChange.Value);
	if (MidiChanMes.ControlChange.Control == g_cv2set)
		g_cc2 = CVCALC(MidiChanMes.ControlChange.Value);
	if (MidiChanMes.ControlChange.Control == g_cv3set)
		g_cc3 = CVCALC(MidiChanMes.ControlChange.Value);

	if (MidiChanMes.ControlChange.Control == g_g0set) {
		if (MidiChanMes.ControlChange.Value < 64)
			GATE0OFF;
		else
			GATE0ON;
	}
	if (MidiChanMes.ControlChange.Control == g_g1set) {
		if (MidiChanMes.ControlChange.Value < 64)
			GATE1OFF;
		else
			GATE1ON;
	}
	if (MidiChanMes.ControlChange.Control == g_g2set) {
		if (MidiChanMes.ControlChange.Value < 64)
			GATE2OFF;
		else
			GATE2ON;
	}
	if (MidiChanMes.ControlChange.Control == g_g3set) {
		if (MidiChanMes.ControlChange.Value < 64)
			GATE3OFF;
		else
			GATE3ON;
	}


	MidiChanMes.Status &= ~fMidiControlChange;
}


#ifdef MCV2313A

static struct _MidiNote *allocN(void)
{
	int i;
	uint16_t msk;

	for (i = 0, msk = 0x01; i < legato_depth; i++, msk<<=1) {
		if (!(MidiNotesTag & msk)) {
			MidiNotesTag |= msk;
			return (&MidiNotes[i]);
		}
	}
	return NULL;
}


static void freeN(struct _MidiNote *addr)
{
	int i;
	uint16_t msk;

	for (i = 0, msk = 0x01; i < legato_depth; i++, msk<<=1) {
		if (addr == &MidiNotes[i]) 
			MidiNotesTag &= ~msk;
	}
}


static void midi_pbend(void)
{
	uint8_t tmp;

	tmp = MidiChanMes.PitchBend.Value>>6;
	g_pitch = BENDCALC(tmp);

	MidiChanMes.Status &= ~fMidiPitchBend;
}


static void midi_noteon(void)
{
	struct _MidiNote *ptr, *newnote;

#ifdef PRIORITY_LATTER
	if ((newnote = allocN()) == NULL) {
		ptr = NoteStart.next;
		NoteStart.next = ptr->next;
		freeN(ptr);
		newnote = allocN();
	}
	newnote->Note = MidiChanMes.NoteOn.Note;
	newnote->Velocity = MidiChanMes.NoteOn.Velocity;
	for (ptr = &NoteStart; ptr->next; ptr = ptr->next) {
		if (ptr->next == &NoteEnd) {
			newnote->next = ptr->next;
			ptr->next = newnote;
			g_note = CVCALC(newnote->Note);
			g_vel = CVCALC(newnote->Velocity);
			tmcnt1m = 77;
			retrigger_time = legato_time;
			g_gate = 1;
			break;
		}
	}
#endif // PRIORITY_LATTER

#ifdef PRIORITY_HIGHER
	if ((newnote = allocN()) != NULL) {
		newnote->Note = MidiChanMes.NoteOn.Note;
		newnote->Velocity = MidiChanMes.NoteOn.Velocity;
		for (ptr = &NoteStart; ptr->next; ptr = ptr->next) {
			if (MidiChanMes.NoteOn.Note > ptr->Note) {
				if (MidiChanMes.NoteOn.Note < ptr->next->Note) {
					if ((ptr->next) == &NoteEnd) {
						// new top note
						g_note = CVCALC(newnote->Note);
						g_vel = CVCALC(newnote->Velocity);
						tmcnt1m = 77;
						retrigger_time = legato_time;
						g_gate = 1;
					}
					newnote->next = ptr->next;
					ptr->next = newnote;
					break;
				}
			}
		}
	}
#endif		// ifdef PRIORITY_HIGHER
	MidiChanMes.Status &= ~fMidiNoteOn;
}

static void midi_noteoff(void)
{
	struct _MidiNote *ptr, *lastptr;

#ifdef PRIORITY_LATTER

	for (ptr = &NoteStart; ptr->next; ptr = ptr->next) {
		if (MidiChanMes.NoteOff.Note == ptr->next->Note) {
			lastptr = ptr->next;
			if ( (lastptr->next == &NoteEnd) && !(ptr == &NoteStart) ) {
				g_note = CVCALC(ptr->Note);
#ifdef PRESERVE_NOTEON_VELOCITY
				g_vel = CVCALC(ptr->Velocity);
#else
				g_vel = CVCALC(MidiChanMes.NoteOff.Velocity);
#endif
				tmcnt1m = 77;
				retrigger_time = legato_time;
			}
			ptr->next = lastptr->next;
			freeN(lastptr);
			break;
		}
	}

#endif // PRIORITY_LATTER


#ifdef PRIORITY_HIGHER
	for (ptr = &NoteStart; ptr->next; ptr = ptr->next) {
		if (MidiChanMes.NoteOff.Note == ptr->next->Note) {
			lastptr = ptr->next;
			if ( ((lastptr->next) == &NoteEnd) && !(ptr == &NoteStart) ) {
				// update top note, but not update for NoteStart
				g_note = CVCALC(ptr->Note);
#ifdef PRESERVE_NOTEON_VELOCITY
				g_vel = CVCALC(ptr->Velocity);
#else
				g_vel = CVCALC(MidiChanMes.NoteOff.Velocity);
#endif
				tmcnt1m = 77;
				retrigger_time = legato_time;
			}
			ptr->next = lastptr->next;
			freeN(lastptr);
			break;
		}
	}
#endif		// ifdef PRIORITY_HIGHER

	// gate off
	if (NoteStart.next == &NoteEnd) {
		g_gate = 0;
		g_vel = CVCALC(MidiChanMes.NoteOff.Velocity);
	}

	MidiChanMes.Status &= ~fMidiNoteOff;
}

#endif // MCV2313A


static void midi_mode(void)
{
	if (MidiChanMes.ModeChange.Mode == ALL_NOTE_OFF) {
		g_gate = 0;
		MidiNotesTag = 0x0000;
		NoteEnd.next = NULL;
		NoteStart.next = &NoteEnd;
	}
	if (MidiChanMes.ModeChange.Mode == RESET_ALL_CONTROLLERS) {
		CV1 = BENDCALC(0x40);
		CV2 = CVCALC(0x00);
		CV3 = CVCALC(0x00);
		GATE1OFF;
		GATE2OFF;
		GATE3OFF;
	}
	MidiChanMes.Status &= ~fMidiModeChange;
}


uint8_t rb;

int main(void)
{
	int8_t c;

	ioinit();
	sei();

	for (;;) {

	    rb = c = uart_getchar();
	    if(c != -1){
			if( MidiSystemMessage() ){
				MidiChannelMessage();
			}
		}

#ifdef NO_USE
		if (MidiSysMes.Status & fMidiActiveSensing)	wdt_reset();
		if (MidiSysMes.Status & fMidiSystemReset) {
			wdt_enable(WDTO_15MS);
			while (1) ;
		}
#endif
		if (MidiChanMes.Status & fMidiModeChange)		midi_mode();
		if (MidiChanMes.Status & fMidiControlChange)	midi_cvgate();
#ifdef MCV2313A
		if (MidiChanMes.Status & fMidiNoteOn)			midi_noteon();
		if (MidiChanMes.Status & fMidiNoteOff)			midi_noteoff();
		if (MidiChanMes.Status & fMidiPitchBend)		midi_pbend();
#endif

#ifdef MCV2313A

		CV0 = g_note;
		CV1 = g_pitch;
		CV2 = (g_expvel) ? g_cc2 : g_vel;
		CV3 = g_cc3;

		if (g_gate & !(retrigger_time))
			GATE0ON;
		else
			GATE0OFF;

#endif // MCV2313A

#ifdef MCV2313C

		CV0 = g_cc0;
		CV1 = g_cc1;
		CV2 = g_cc2;
		CV3 = g_cc3;

#endif // MCV2313C


	}



	return 0;
}
